#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2020, 2024-2025 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import logging
import subprocess

from diffoscope.tools import tool_required


from .utils.file import File
from .utils.archive import Archive

logger = logging.getLogger(__name__)


class LzmaContainer(Archive):
    def open_archive(self):
        return self

    def close_archive(self):
        pass

    def get_member_names(self):
        return [self.get_compressed_content_name(".lzma")]

    @tool_required("xz")
    def extract(self, member_name, dest_dir):
        dest_path = self.get_path_name(dest_dir)
        logger.debug("lzma extracting to %s", dest_path)
        with open(dest_path, "wb") as fp:
            subprocess.check_call(
                [
                    "xz",
                    "--decompress",
                    "--format=lzma",
                    "--stdout",
                    self.source.path,
                ],
                stdout=fp,
                stderr=None,
            )
        return dest_path


class LzmaFile(File):
    DESCRIPTION = "LZMA compressed files"
    CONTAINER_CLASSES = [LzmaContainer]
    FILE_TYPE_RE = re.compile(r"^LZMA compressed data\b")



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2020 Conrad Ratschan <ratschance@gmail.com>
# Copyright © 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os.path
import re
import subprocess

from .utils.file import File
from .utils.archive import Archive

logger = logging.getLogger(__name__)


class UimageContainer(Archive):
    def open_archive(self):
        return self

    def close_archive(self):
        pass

    def get_member_names(self):
        return [self.get_compressed_content_name(".uboot")]

    def extract(self, member_name, dest_dir):
        dest_path = os.path.join(dest_dir, member_name)
        logger.debug("uImage extracting to %s", dest_path)
        with open(dest_path, "wb") as fp:
            # Use tail to cut off the first 64 bytes without having to do a byte by byte copy in python
            subprocess.check_call(
                ["tail", "-c+65", self.source.path],
                shell=False,
                stdout=fp,
                stderr=None,
            )
        return dest_path


class UimageFile(File):
    """
    U-Boot legacy image files are standard OS boot files (kernel, rootfs) with a 64
    byte header appended to the front of the file by mkimage to provide the load
    address, entrypoint and a CRC for the contained image. By treating this as a
    container and setting our extract method to just cut the first 64 bytes off, the
    rest of diffoscope's tooling can be used to view the differences of the contained
    files.
    """

    DESCRIPTION = "U-Boot legacy image files"
    CONTAINER_CLASSES = [UimageContainer]
    FILE_TYPE_RE = re.compile(r"^u-boot legacy uImage\b")



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2024 Jelle van der Waa <jelle@archlinux.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.


import subprocess
import re
import logging

from diffoscope.tools import tool_required, tool_check_installed
from diffoscope.difference import Difference

from .utils.file import File
from .utils.command import Command, our_check_output

logger = logging.getLogger(__name__)


class UKIDump(Command):
    @tool_required("ukify")
    def cmdline(self):
        return ["ukify", "inspect", self.path]


class UKIFile(File):
    DESCRIPTION = "UKI image (.efi)"
    FILE_EXTENSION_SUFFIX = {".efi"}
    FILE_TYPE_RE = re.compile(r"^PE32\+")

    @classmethod
    def recognizes(cls, file):
        if not super().recognizes(file):
            return False

        if not tool_check_installed("objdump"):
            return False

        try:
            our_check_output(
                ["objdump", "-h", "--section", ".linux", file.path],
            )
            return True
        except subprocess.SubprocessError:
            return False

    def compare_details(self, other, source=None):
        return [Difference.from_operation(UKIDump, self.path, other.path)]



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017, 2019-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import json
from collections import OrderedDict

from .utils import Presenter

JSON_FORMAT_VERSION = 1
JSON_FORMAT_MAGIC = "diffoscope-json-version"


class JSONPresenter(Presenter):
    def __init__(self, print_func):
        self.stack = []
        self.print_func = print_func

        super().__init__()

    def start(self, difference):
        root = []
        self.stack = [root]
        super().start(difference)

        root[0][JSON_FORMAT_MAGIC] = JSON_FORMAT_VERSION
        root[0].move_to_end(JSON_FORMAT_MAGIC, last=False)
        self.print_func(json.dumps(root[0]))

    def visit_difference(self, difference):
        while self.depth + 1 < len(self.stack):
            self.stack.pop()

        elements = [
            ("source1", difference.source1),
            ("source2", difference.source2),
        ]
        if difference.comments:
            elements += [("comments", [x for x in difference.comments])]
        if difference.has_internal_linenos:
            elements += [("has_internal_linenos", True)]
        elements += [("unified_diff", difference.unified_diff)]

        child_differences = []
        if difference.details:
            elements += [("details", child_differences)]

        self.stack[-1].append(OrderedDict(elements))
        if difference.details:
            self.stack.append(child_differences)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2017-2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import io
import os
import errno
import fcntl
import hashlib
import logging
import itertools
import threading
import subprocess

from difflib import Differ
from multiprocessing.dummy import Queue

from .tools import get_tool_name, tool_required
from .config import Config
from .profiling import profile
from .tempfiles import get_temporary_directory

DIFF_CHUNK = 4096

logger = logging.getLogger(__name__)
re_diff_change = re.compile(r"^([+-@]).*", re.MULTILINE)


class DiffParser:
    RANGE_RE = re.compile(
        # example: '@@ -26814,9 +26814,8 @@'
        rb"""
          ^
            @@\s+
              -
              (?P<start1>\d+)(,(?P<len1>\d+))?
            \s+
              \+
              (?P<start2>\d+)(,(?P<len2>\d+))?
            \s+@@
          $
        """,
        re.VERBOSE,
    )

    def __init__(self, output, end_nl_q1, end_nl_q2):
        self._output = output
        self._end_nl_q1 = end_nl_q1
        self._end_nl_q2 = end_nl_q2
        self._action = self.read_headers
        self._diff = io.BytesIO()
        self._success = False
        self._remaining_hunk_lines = None
        self._block_len = None
        self._direction = None
        self._end_nl = None
        self._max_lines = Config().max_diff_block_lines_saved

    @property
    def diff(self):
        return self._diff.getvalue().decode("UTF-8", errors="replace")

    @property
    def success(self):
        return self._success

    def parse(self):
        for line in self._output.split(b"\n"):
            self._action = self._action(line)

        self._action = None
        self._success = True

    def read_headers(self, line):
        if not line:
            return None

        if line.startswith(b"---"):
            return self.read_headers

        if line.startswith(b"+++"):
            return self.read_headers

        found = DiffParser.RANGE_RE.match(line)

        if not found:
            raise ValueError(f"Unable to parse diff headers: {line!r}")

        self._diff.write(line + b"\n")
        if found.group("len1"):
            self._remaining_hunk_lines = int(found.group("len1"))
        else:
            self._remaining_hunk_lines = 1
        if found.group("len2"):
            self._remaining_hunk_lines += int(found.group("len2"))
        else:
            self._remaining_hunk_lines += 1

        self._direction = None

        return self.read_hunk

    def read_hunk(self, line):
        if not line:
            return None

        if line[:1] == b" ":
            self._remaining_hunk_lines -= 2
        elif line[:1] == b"+":
            self._remaining_hunk_lines -= 1
        elif line[:1] == b"-":
            self._remaining_hunk_lines -= 1
        elif line[:1] == b"\\":
            # When both files don't end with \n, do not show it as a difference
            if self._end_nl is None:
                end_nl1 = self._end_nl_q1.get()
                end_nl2 = self._end_nl_q2.get()
                self._end_nl = end_nl1 and end_nl2
            if not self._end_nl:
                return self.read_hunk
        elif self._remaining_hunk_lines == 0:
            return self.read_headers(line)
        else:
            raise ValueError(f"Unable to parse diff hunk: {line!r}")

        self._diff.write(line + b"\n")

        if line[:1] in (b"-", b"+"):
            if line[:1] == self._direction:
                self._block_len += 1
            else:
                self._block_len = 1
                self._direction = line[:1]

            if self._block_len >= self._max_lines:
                return self.skip_block
        else:
            self._block_len = 1
            self._direction = line[:1]

        return self.read_hunk

    def skip_block(self, line):
        if self._remaining_hunk_lines == 0 or line[:1] != self._direction:
            removed = self._block_len - Config().max_diff_block_lines_saved
            if removed:
                self._diff.write(
                    b"%s[ %d lines removed ]\n" % (self._direction, removed)
                )
            return self.read_hunk(line)

        self._block_len += 1
        self._remaining_hunk_lines -= 1

        return self.skip_block


@tool_required("diff")
def run_diff(fifo1, fifo2, end_nl_q1, end_nl_q2):
    cmd = [
        get_tool_name("diff"),
        "-a",
        "-U" + str(Config().diff_context),
        fifo1,
        fifo2,
    ]

    logger.debug("Running %s", " ".join(cmd))

    p = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    parser = DiffParser(p.stdout, end_nl_q1, end_nl_q2)
    parser.parse()

    logger.debug(
        "%s: returncode %d, parsed %s",
        " ".join(cmd),
        p.returncode,
        parser.success,
    )

    if not parser.success and p.returncode not in (0, 1):
        raise subprocess.CalledProcessError(
            p.returncode, cmd, output=parser.diff.encode("utf-8")
        )

    if p.returncode == 0:
        return None

    return parser.diff


class FIFOFeeder(threading.Thread):
    def __init__(self, feeder, fifo_path, end_nl_q=None, daemon=True, *args):
        os.mkfifo(fifo_path)
        super().__init__(daemon=daemon)
        self.feeder = feeder
        self.fifo_path = fifo_path
        self.end_nl_q = Queue() if end_nl_q is None else end_nl_q
        self._exception = None
        self._want_join = threading.Event()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        os.remove(self.fifo_path)
        self.join()

    def run(self):
        try:
            # Try to open the FIFO nonblocking, so we can periodically check
            # if the main thread wants us to wind down.  If it does, there's no
            # more need for the FIFO, so stop the thread.
            while True:
                try:
                    fd = os.open(self.fifo_path, os.O_WRONLY | os.O_NONBLOCK)
                except OSError as e:
                    if e.errno != errno.ENXIO:
                        raise
                    if self._want_join.is_set():
                        return
                else:
                    break

            # Now clear the fd's nonblocking flag to let writes block normally.
            fcntl.fcntl(fd, fcntl.F_SETFL, 0)

            with open(fd, "wb") as fifo:
                # The queue works around a unified diff limitation: if there's
                # no newlines in both don't make it a difference
                end_nl = self.feeder(fifo)
                self.end_nl_q.put(end_nl)
        except Exception as error:
            self._exception = error

    def join(self):
        self._want_join.set()
        super().join()
        if self._exception is not None:
            raise self._exception


class _Feeder:
    """
    A 'feeder' is a specialized writer.

    A 'feeder' is a callable that takes as argument a writeable file, and
    writes to it.  Feeders can transform the written data, truncate it,
    checksum it, and so on.  The callable must return True to represent that
    the data had a terminating newline, and False otherwise.

    Feeders are created by the functions make_feeder_from_raw_reader() and
    empty_file_feeder().  The returned objects are closures, and are not
    (currently?) instances of any particular class.
    """


def empty_file_feeder():
    """
    Returns a feeder that simulates an empty file.

    See _Feeder for feeders.
    """

    def feeder(f):
        return False

    return feeder


def make_feeder_from_raw_reader(in_file, filterfn=None):
    """
    Create a feeder that checksums, truncates, and transcodes the data.  The
    optional argument filterfn is a callable that gets passed each line, and
    returns the line that should be used in its stead.  (There is no facility
    for filterfn to discard a line entirely.)

    See _Feeder for feeders.
    """

    def feeder(out_file):
        h = None
        end_nl = False
        max_lines = Config().max_diff_input_lines
        line_count = 0

        if max_lines < float("inf"):
            h = hashlib.sha256()

        for buf in in_file:
            line_count += 1
            out = filterfn(buf) if filterfn else buf
            if h:
                h.update(out)
            if line_count < max_lines:
                out_file.write(out)
            end_nl = buf[-1] == "\n"

        if h and line_count >= max_lines:
            out_file.write(
                "[ Too much input for diff (SHA: {}) ]\n".format(
                    h.hexdigest()
                ).encode("utf-8")
            )
            end_nl = True

        return end_nl

    return feeder


def diff(feeder1, feeder2):
    with get_temporary_directory() as tmpdir:
        fifo1_path = os.path.join(tmpdir, "fifo1")
        fifo2_path = os.path.join(tmpdir, "fifo2")
        with FIFOFeeder(feeder1, fifo1_path) as fifo1, FIFOFeeder(
            feeder2, fifo2_path
        ) as fifo2:
            return run_diff(
                fifo1_path, fifo2_path, fifo1.end_nl_q, fifo2.end_nl_q
            )


def diff_split_lines(diff, keepends=True):
    lines = diff.split("\n")
    if not keepends:
        return lines
    return [line + "\n" for line in lines[:-1]] + (
        [lines[-1]] if lines[-1] else []
    )


def reverse_unified_diff(diff):
    assert isinstance(diff, str)
    res = []
    # XXX - should move to just use plain bytes nearly everywhere until ready
    # to print instead of using strings internally as well.
    for line in diff_split_lines(diff):
        line = line.encode("utf-8", errors="replace")
        assert isinstance(line, bytes)
        found = DiffParser.RANGE_RE.match(line)

        if found:
            before = found.group("start2")
            if found.group("len2") is not None:
                before += b"," + found.group("len2")

            after = found.group("start1")
            if found.group("len1") is not None:
                after += b"," + found.group("len1")

            res.append(b"@@ -%s +%s @@\n" % (before, after))
        elif line.startswith(b"-"):
            res.append(b"+")
            res.append(line[1:])
        elif line.startswith(b"+"):
            res.append(b"-")
            res.append(line[1:])
        else:
            res.append(line)
    return b"".join(res).decode("utf-8", errors="replace")


def color_unified_diff(diff):
    RESET = "\033[0m"
    RED, GREEN, CYAN = "\033[31m", "\033[32m", "\033[0;36m"

    def repl(m):
        return "{}{}{}".format(
            {"-": RED, "@": CYAN, "+": GREEN}[m.group(1)], m.group(0), RESET
        )

    return re_diff_change.sub(repl, diff)


DIFFON = "\x01"
DIFFOFF = "\x02"
MAX_WAGNER_FISCHER_SIZE = (
    256  # any higher, and linediff takes >1 second and >200MB RAM
)


def _linediff_sane(x):
    # turn non-printable chars into "."
    return "." if ord(x) < 32 and x not in "\t\n" else x


def diffinput_truncate(s, sz):
    # Truncate, preserving uniqueness
    if len(s) > sz:
        s = s[
            :sz
        ] + "[ ... truncated by diffoscope; len: {}, SHA: {} ... ]".format(
            len(s[sz:]), hashlib.sha256(s[sz:].encode("utf-8")).hexdigest()
        )
    return s


def _commonprefix(s, t):
    # calculate the common prefix of two strings
    s1 = min(s, t)
    s2 = max(s, t)
    for i, c in enumerate(s1):
        if c != s2[i]:
            return s1[:i]
    return s1


def linediff(s, t, diffon, diffoff):
    # calculate common prefix/suffix, easy optimisation for Wagner-Fischer
    prefix = _commonprefix(s, t)
    if prefix:
        s = s[len(prefix) :]
        t = t[len(prefix) :]
    suffix = _commonprefix(s[::-1], t[::-1])[::-1]
    if suffix:
        s = s[: -len(suffix)]
        t = t[: -len(suffix)]

    # truncate so Wagner-Fischer doesn't blow up RAM
    s = diffinput_truncate(s, MAX_WAGNER_FISCHER_SIZE)
    t = diffinput_truncate(t, MAX_WAGNER_FISCHER_SIZE)
    l1, l2 = zip(*linediff_simplify(linediff_wagnerfischer(s, t)))

    def to_string(k, v):
        sanev = "".join(_linediff_sane(c) for c in v)
        return (diffon + sanev + diffoff) if k else sanev

    s1 = "".join(to_string(*p) for p in l1)
    t1 = "".join(to_string(*p) for p in l2)

    return f"{prefix}{s1}{suffix}", f"{prefix}{t1}{suffix}"


def linediff_wagnerfischer(s, t):
    """
    Line diff algorithm, originally from diff2html.

    https://en.wikipedia.org/wiki/Wagner%E2%80%93Fischer_algorithm

    Finds the minimum (levenshtein) edit distance between two strings, but has
    quadratic performance O(m*n).
    """
    m, n = len(s), len(t)
    d = [[(0, 0) for i in range(n + 1)] for i in range(m + 1)]

    d[0][0] = (0, (0, 0))
    for i in range(1, m + 1):
        d[i][0] = (i, (i - 1, 0))
    for j in range(1, n + 1):
        d[0][j] = (j, (0, j - 1))

    # NB. This loop is O(len(s) * len(t))
    for i in range(1, m + 1):
        for j in range(1, n + 1):
            if s[i - 1] == t[j - 1]:
                cost = 0
            else:
                cost = 1
            d[i][j] = min(
                (d[i - 1][j][0] + 1, (i - 1, j)),
                (d[i][j - 1][0] + 1, (i, j - 1)),
                (d[i - 1][j - 1][0] + cost, (i - 1, j - 1)),
            )

    coords = []
    coord = (m, n)
    while coord != (0, 0):
        coords.insert(0, coord)
        x, y = coord
        coord = d[x][y][1]

    for coord in coords:
        cx, cy = coord
        child_val = d[cx][cy][0]

        father_coord = d[cx][cy][1]
        fx, fy = father_coord
        father_val = d[fx][fy][0]

        diff = (cx - fx, cy - fy)

        if diff == (0, 1):
            yield (False, ""), (True, t[fy])
        elif diff == (1, 0):
            yield (True, s[fx]), (False, "")
        elif child_val - father_val == 1:
            yield (True, s[fx]), (True, t[fy])
        else:
            assert s[fx] == t[fy]
            yield (False, s[fx]), (False, t[fy])


def linediff_simplify(g):
    """Simplify the output of Wagner-Fischer."""
    current = None
    for l, r in g:
        if not current:
            current = l, r
        elif current[0][0] == l[0] and current[1][0] == r[0]:
            current = (
                (l[0], current[0][1] + l[1]),
                (r[0], current[1][1] + r[1]),
            )
        else:
            yield current
            current = l, r
    if current:
        yield current


class SideBySideDiff:
    """Calculates a side-by-side diff from a unified diff."""

    def __init__(self, unified_diff, diffon=DIFFON, diffoff=DIFFOFF):
        self.unified_diff = unified_diff
        self.diffon = diffon
        self.diffoff = diffoff
        self.reset()

    def reset(self):
        self.differ = Differ()
        self.buf = []
        self.add_cpt = 0
        self.del_cpt = 0
        self.line1 = 0
        self.line2 = 0
        self.hunk_off1 = 0
        self.hunk_size1 = 0
        self.hunk_off2 = 0
        self.hunk_size2 = 0
        self._bytes_processed = 0

    @property
    def bytes_processed(self):
        return self._bytes_processed

    def match_lines(self, l0, l1):
        """
        Given two lists of lines, use python's difflib to make the diff a bit
        "smarter", as it better detects added lines (see !64)
        """

        if len(l0) + len(l1) > 750:
            # difflib.Differ.compare is at least O(n^2), so don't call it if
            # our inputs are too large.
            yield "C", "Diff chunk too large, falling back to line-by-line diff ({} lines added, {} lines removed)".format(
                self.add_cpt, self.del_cpt
            )
            for line0, line1 in itertools.zip_longest(l0, l1, fillvalue=""):
                yield from self.yield_line(line0, line1)
            return

        saved_line = None
        for line in self.differ.compare(l0, l1):
            # Skip lines with details as they don't matter to us
            if line.startswith("? "):
                continue

            # When we encounter a +, it may either be because a line was
            # added, or because a line was modified (in that case,
            # save_line will not be None)
            if line.startswith("+ "):
                yield from self.yield_line(saved_line or "", line[2:])
                saved_line = None
                continue

            # If saved_line is not None, we need to yield from it before
            # moving on
            if saved_line is not None:
                yield from self.yield_line(saved_line, "")
                saved_line = None

            # Handle removed and unchanged lines
            if line.startswith("- "):
                saved_line = line[2:]
            elif line.startswith("  "):
                line = line[2:]
                yield from self.yield_line(line, line)

        # Make sure we don't forget a line
        if saved_line is not None:
            yield from self.yield_line(saved_line, "")

    def empty_buffer(self):
        if self.del_cpt == 0 or self.add_cpt == 0:
            # All lines are unchanged lines
            for l in self.buf:
                yield from self.yield_line(l[0], l[1])
        elif self.del_cpt != 0 and self.add_cpt != 0:
            l0 = [pair[0] for pair in self.buf if pair[0] is not None]
            l1 = [pair[1] for pair in self.buf if pair[1] is not None]
            yield from self.match_lines(l0, l1)

    def yield_line(self, s1, s2):
        orig1 = s1
        orig2 = s2

        if s1 is None and s2 is None:
            type_name = "unmodified"
        elif s1 == "" and s2 == "":
            type_name = "unmodified"
        elif s1 is None or s1 == "":
            type_name = "added"
        elif s2 is None or s2 == "":
            type_name = "deleted"
        elif (
            orig1 == orig2
            and not s1.endswith("lines removed ]")
            and not s2.endswith("lines removed ]")
        ):
            type_name = "unmodified"
        else:
            type_name = "changed"
            with profile("diff", "linediff"):
                s1, s2 = linediff(s1, s2, self.diffon, self.diffoff)

        yield "L", (type_name, s1, self.line1, s2, self.line2)

        m = orig1 and re.match(r"^\[ (\d+) lines removed \]$", orig1)
        if m:
            self.line1 += int(m.group(1))
        elif orig1:
            self.line1 += 1
        m = orig2 and re.match(r"^\[ (\d+) lines removed \]$", orig2)
        if m:
            self.line2 += int(m.group(1))
        elif orig2:
            self.line2 += 1

        self.add_cpt = 0
        self.del_cpt = 0
        self.buf = []

    def items(self):
        """Yield the items that form the side-by-side diff.

        Each item is a (type, value) tuple, as follows:

        type == "H", value is a tuple representing a hunk header
            hunk_offset1, hunk_size1, hunk_offset2, hunk_size2 = value
            all ints

        type == "L", value is a tuple representing a line of a hunk
            mode, line1, lineno1, line2, lineno2 = value
            where mode is one of {"unmodified", "added", "deleted", "changed"}
            line* are strings
            lineno* are ints

        type == "C", value is a comment
            comment = value
            a string
        """
        self.reset()

        in_header = True
        for l in diff_split_lines(self.unified_diff, False):
            self._bytes_processed += len(l) + 1
            m = re.match(r"^--- ([^\s]*)", l)
            if m and in_header:
                yield from self.empty_buffer()
                continue
            m = re.match(r"^\+\+\+ ([^\s]*)", l)
            if m and in_header:
                yield from self.empty_buffer()
                continue

            m = re.match(r"@@ -(\d+),?(\d*) \+(\d+),?(\d*)", l)
            if m:
                in_header = False
                yield from self.empty_buffer()
                hunk_data = map(lambda x: x == "" and 1 or int(x), m.groups())
                (
                    self.hunk_off1,
                    self.hunk_size1,
                    self.hunk_off2,
                    self.hunk_size2,
                ) = hunk_data
                self.line1, self.line2 = self.hunk_off1, self.hunk_off2
                yield "H", (
                    self.hunk_off1,
                    self.hunk_size1,
                    self.hunk_off2,
                    self.hunk_size2,
                )
                continue

            if re.match(r"^\[", l):
                yield from self.empty_buffer()
                yield "C", l

            if re.match(r"^\\ No newline", l):
                if self.hunk_size2 == 0:
                    self.buf[-1] = (
                        self.buf[-1][0],
                        self.buf[-1][1] + "\n" + l[2:],
                    )
                else:
                    self.buf[-1] = (
                        self.buf[-1][0] + "\n" + l[2:],
                        self.buf[-1][1],
                    )
                continue

            if self.hunk_size1 <= 0 and self.hunk_size2 <= 0:
                yield from self.empty_buffer()
                continue

            m = re.match(r"^\+\[ (\d+) lines removed \]$", l)
            if m:
                self.add_cpt += int(m.group(1))
                self.hunk_size2 -= int(m.group(1))
                self.buf.append((None, l[1:]))
                continue

            if re.match(r"^\+", l):
                self.add_cpt += 1
                self.hunk_size2 -= 1
                self.buf.append((None, l[1:]))
                continue

            m = re.match(r"^-\[ (\d+) lines removed \]$", l)
            if m:
                self.del_cpt += int(m.group(1))
                self.hunk_size1 -= int(m.group(1))
                self.buf.append((l[1:], None))
                continue

            if re.match(r"^-", l):
                self.del_cpt += 1
                self.hunk_size1 -= 1
                self.buf.append((l[1:], None))
                continue

            if re.match(r"^ ", l) and self.hunk_size1 and self.hunk_size2:
                yield from self.empty_buffer()
                self.hunk_size1 -= 1
                self.hunk_size2 -= 1
                self.buf.append((l[1:], l[1:]))
                continue

            yield from self.empty_buffer()

        yield from self.empty_buffer()



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2024 Seth Michael Larson <seth@python.org>
# Copyright © 2024-2025 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.
import hashlib
import re
import struct
import xml.etree.ElementTree as ET
import zlib
import os

from .utils.file import File
from .utils.archive import Archive
from diffoscope.difference import Difference


class XarContainer(Archive):
    def get_member_names(self):
        toc_xml = self.parse_toc_xml()
        for file_tag in toc_xml.iter("file"):
            yield file_tag.get("id")  # Use IDs instead of names for members.

    def open_archive(self):
        pass

    def close_archive(self):
        pass

    def extract(self, member_name, dest_dir):
        toc_xml = self.parse_toc_xml()

        # Find the file name and data heap offset.
        for file_tag in toc_xml.iter("file"):
            if file_tag.get("id") == member_name:
                file_name = file_tag.find(".//name").text
                data_offset = int(file_tag.find(".//data/offset").text)
                data_length = int(file_tag.find(".//data/length").text)
                break
        else:
            raise KeyError(member_name)

        # Write data from the heap into the temporary directory.
        # We automatically handle gzipped data thanks to the header.
        dest_path = os.path.join(dest_dir, file_name)
        with open(dest_path, mode="wb") as fw, open(
            self._source.path, mode="rb"
        ) as fr:
            fr.seek(self._heap_offset + data_offset, 0)
            fw.write(fr.read(data_length))

        return dest_path

    def parse_toc_xml(self):
        if getattr(self, "_toc_xml", None) is None:
            with open(self._source.path, mode="rb") as f:
                # Skip the magic and format, we're looking for
                # header and TOC compressed lengths.
                (
                    header_length,
                    toc_compressed_length,
                ) = struct.unpack(">xxxxHxxQ", f.read(16))

                # Read, decompress, and parse the TOC as XML. Save heap offset for later.
                f.seek(header_length, 0)
                toc_bytes = f.read(toc_compressed_length)
                toc_as_text = zlib.decompress(toc_bytes).decode("utf-8")
                self._toc_xml = ET.XML(toc_as_text)
                self._heap_offset = header_length + toc_compressed_length

        return self._toc_xml


class XarFile(File):
    DESCRIPTION = "eXtensible ARchive files"
    CONTAINER_CLASSES = [XarContainer]
    FILE_TYPE_RE = re.compile(r"\bpkg\b")
    FALLBACK_FILE_EXTENSION_SUFFIX = {
        ".xar",
        ".pkg",
    }  # NOTE: Facebook's Executable Archive format also uses '.xar'.
    FALLBACK_FILE_TYPE_HEADER_PREFIX = b"xar!"

    def compare_details(self, other, source=None):
        self_xar_header, self_xar_toc = describe_xar(self.path)
        other_xar_header, other_xar_toc = describe_xar(other.path)
        return [
            Difference.from_text(
                self_xar_header,
                other_xar_header,
                self.path,
                other.path,
                source="XAR Header",
            ),
            Difference.from_text(
                self_xar_toc,
                other_xar_toc,
                self.path,
                other.path,
                source="XAR Table of Contents",
            ),
        ]


def describe_xar(path):
    with open(path, mode="rb") as f:
        magic = f.read(4)

        # Read the fixed portion of the XAR header
        # Padding length is calculated using header length.
        (
            header_length,
            format_version,
            toc_compressed_length,
            toc_uncompressed_length,
            checksum_alg,
        ) = struct.unpack(">HHQQI", f.read(24))

        known_checksum_algs = {
            0: "NONE",
            1: "SHA1",
            2: "MD5",
            3: "SHA-256",
            4: "SHA-512",
        }

        header_lines = [
            "magic:                   {}".format(magic),
            "format version:          {}".format(format_version),
            "TOC compressed length:   {}".format(toc_compressed_length),
            "TOC uncompressed length: {}".format(toc_uncompressed_length),
            "checksum:                {} ({})".format(
                checksum_alg, known_checksum_algs.get(checksum_alg, "???")
            ),
        ]

        # Note that this 'header length' includes the 4 bytes of magic, hence 28.
        padding_length = header_length - 28
        if padding_length > 0:  # Padding is optional.
            padding = f.read(padding_length)
            header_lines.append("padding:                 {}".format(padding))

        # Read the TOC which is always DEFLATE compressed.
        toc_bytes = f.read(toc_compressed_length)
        toc_as_text = zlib.decompress(toc_bytes).decode("utf-8")

        # Read the entire heap and add properties that allow detecting
        # "invisible" differences in the heap, for example if data is inserted
        # but isn't referenced in the TOC. This shouldn't happen in a normal XAR file.
        heap_bytes = f.read()
        header_lines.extend(
            [
                "heap length:             {}".format(len(heap_bytes)),
                "heap checksum:           {}".format(
                    hashlib.sha256(heap_bytes).hexdigest()
                ),
            ]
        )

        header_as_text = "\n".join(header_lines)
        return header_as_text, toc_as_text



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016 Reiner Herrmann <reiner@reiner-h.de>
# Copyright © 2016-2022, 2025-2026 Chris Lamb <lamby@debian.org>
# Copyright © 2022 FC Stegerman <flx@obfusk.net>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import binascii
import textwrap
import os.path
import logging
import itertools
import subprocess

from diffoscope.difference import Difference
from diffoscope.exc import RequiredToolNotFound
from diffoscope.tools import (
    tool_required,
    find_executable,
    python_module_missing,
)
from diffoscope.tempfiles import get_temporary_directory

from .text import TextFile
from .utils.archive import Archive, ArchiveMember
from .utils.command import Command
from .utils.compare import compare_files
from .utils.specialize import specialize_as
from .zip import ZipContainer, zipinfo_differences, ZipFileBase
from .missing_file import MissingFile

logger = logging.getLogger(__name__)

try:
    import androguard
except ImportError:
    python_module_missing("androguard")
    androguard = None


class ApkContainer(Archive):
    @property
    def path(self):
        return self._path

    @tool_required("apktool")
    @tool_required("zipinfo")
    def open_archive(self):
        self._members = []
        self._tmpdir = get_temporary_directory(suffix="apk")
        self._andmanifest = None
        self._andmanifest_orig = None

        logger.debug(
            "Extracting %s to %s", self.source.name, self._tmpdir.name
        )

        subprocess.check_call(
            (
                "apktool",
                "decode",
                "--force",
                "--keep-broken-res",
                "--match-original",
                "--output",
                self._tmpdir.name,
                self.source.path,
            ),
            stderr=None,
            stdout=subprocess.PIPE,
        )

        # Optionally extract a few files that apktool does not
        for x in ("classes.dex", "resources.arsc"):
            subprocess.call(
                ("unzip", "-d", self._tmpdir.name, self.source.path, x),
                stderr=subprocess.PIPE,
                stdout=subprocess.PIPE,
            )

        # ... including "classes2.dex", "classes3.dex", etc.
        for x in itertools.count(2):
            try:
                subprocess.check_call(
                    (
                        "unzip",
                        "-d",
                        self._tmpdir.name,
                        self.source.path,
                        f"classes{x}.dex",
                    ),
                    stderr=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                )
            except subprocess.CalledProcessError:
                break

        for root, _, files in os.walk(self._tmpdir.name):
            current_dir = []

            for filename in files:
                abspath = os.path.join(root, filename)

                # apktool.yml is a file created by apktool which contains
                # metadata information. We eename it for clarity and always
                # make it appear at the beginning of the directory listing for
                # reproducibility.
                if filename == "apktool.yml":
                    abspath = filter_apk_metadata(
                        abspath, os.path.basename(self.source.path)
                    )
                    relpath = abspath[len(self._tmpdir.name) + 1 :]
                    current_dir.insert(0, relpath)
                    continue

                relpath = abspath[len(self._tmpdir.name) + 1 :]

                if filename == "AndroidManifest.xml":
                    containing_dir = root[len(self._tmpdir.name) + 1 :]
                    if containing_dir == "original":
                        self._andmanifest_orig = relpath
                    if containing_dir == "":
                        self._andmanifest = relpath
                    continue

                current_dir.append(relpath)

            self._members.extend(current_dir)

        return self

    def get_android_manifest(self):
        return (
            self.get_member(self._andmanifest) if self._andmanifest else None
        )

    def get_original_android_manifest(self):
        if self._andmanifest_orig:
            return self.get_member(self._andmanifest_orig)
        return MissingFile("/dev/null", self._andmanifest_orig)

    def close_archive(self):
        pass

    def get_member_names(self):
        return self._members

    def get_member(self, member_name):
        member = ArchiveMember(self, member_name)
        if member_name.endswith(".smali") and member_name.startswith("smali"):
            # smali{,_classesN}/**/*.smali files from apktool are always text,
            # and using libmagic on thousands of these files takes minutes
            return specialize_as(TextFile, member)
        return member

    def extract(self, member_name, dest_dir):
        return os.path.join(self._tmpdir.name, member_name)

    def compare_manifests(self, other):
        my_android_manifest = self.get_android_manifest()
        other_android_manifest = other.get_android_manifest()
        comment = None
        diff_manifests = None
        if my_android_manifest and other_android_manifest:
            source = "AndroidManifest.xml (decoded)"
            diff_manifests = compare_files(
                my_android_manifest, other_android_manifest, source=source
            )
            if diff_manifests is None:
                comment = "No difference found for decoded AndroidManifest.xml"
        else:
            comment = (
                "No decoded AndroidManifest.xml found "
                + "for one of the APK files."
            )
        if diff_manifests:
            return diff_manifests

        source = "AndroidManifest.xml (original / undecoded)"
        diff_manifests = compare_files(
            self.get_original_android_manifest(),
            other.get_original_android_manifest(),
            source=source,
        )
        if diff_manifests is not None:
            diff_manifests.add_comment(comment)
        return diff_manifests

    def compare(self, other, *args, **kwargs):
        differences = []
        try:
            differences.append(self.compare_manifests(other))
        except AttributeError:  # no apk-specific methods, e.g. MissingArchive
            pass
        differences.extend(super().compare(other, *args, **kwargs))
        return differences


class Apksigner(Command):
    VALID_RETURNCODES = {0, 1}

    @tool_required("apksigner")
    def cmdline(self):
        # Older versions of the `apksigner` binary under /usr/bin (or similar)
        # are a symbolic link to the apksigner .jar file. If we detect a .jar
        # we resolve its 'real' location and pass that to `java -jar`, so we
        # don't need kernel binfmt_misc to execute. We can't do this in all
        # situations as later versions of apksigner use a wrapper script and
        # will therefore fail to run at all if we use `java -jar`.
        apksigner = os.path.realpath(find_executable("apksigner"))

        prefix = ["java", "-jar"] if apksigner.endswith(".jar") else []

        return prefix + [
            apksigner,
            "verify",
            "--verbose",
            "--print-certs",
            self.path,
        ]


class ApkFile(ZipFileBase):
    DESCRIPTION = "Android APK files"
    FILE_TYPE_HEADER_PREFIX = b"PK\x03\x04"
    FILE_TYPE_RE = re.compile(r"^(Android package|(Java|Zip) archive data)\b")
    FILE_EXTENSION_SUFFIX = {".apk", ".aar"}
    CONTAINER_CLASSES = [ApkContainer, ZipContainer]

    @property
    def as_container(self):
        # If we found no differences before the APK Signing Block we return None
        # here to prevent apktool from being run needlessly (which can take up a
        # significant amount of extra time) via ApkContainer (since there's no
        # API that allows us to selectively disable use of container classes in
        # cases like these).
        if getattr(self, "_disable_container_compare", False):
            return None  # don't run apktool
        return super().as_container

    def compare_details(self, other, source=None):
        self.check_differences_before_signing_block(other)

        differences = zipinfo_differences(self, other)

        try:
            x = Difference.from_operation(Apksigner, self.path, other.path)
            if x is not None:
                differences.insert(0, x)
        except RequiredToolNotFound as exc:  # noqa
            # Don't require apksigner
            self.add_comment(exc.get_comment())

        if androguard is None:
            self.add_comment(
                "'androguard' Python package not installed; cannot extract V2 signing keys."
            )
        else:
            x = Difference.from_text_readers(
                get_v2_signing_keys(self.path),
                get_v2_signing_keys(other.path),
                self.path,
                other.path,
                source="APK Signing Block",
            )
            if x is not None:
                differences.insert(0, x)

        return differences

    def check_differences_before_signing_block(self, other):
        try:
            self._check_differences_before_signing_block(other)
        except (RequiredToolNotFound, ImportError):
            self.add_comment(
                "'apksigcopier' Python package not installed; unconditionally running 'apktool'."
            )
            return

    @tool_required("apksigcopier")
    def _check_differences_before_signing_block(self, other):
        import apksigcopier

        try:
            offset_self, _ = apksigcopier.extract_v2_sig(self.path)
            offset_other, _ = apksigcopier.extract_v2_sig(other.path)
        except Exception:
            return

        if offset_self != offset_other:
            return

        with open(self.path, "rb") as fh_self:
            with open(other.path, "rb") as fh_other:
                while fh_self.tell() < offset_self:
                    size = min(offset_self - fh_self.tell(), 4096)
                    if fh_self.read(size) != fh_other.read(size):
                        return

        self.add_comment(
            "No differences before APK Signing Block; not running 'apktool'."
        )

        self._disable_container_compare = True
        other._disable_container_compare = True


def get_v2_signing_keys(path):
    try:
        from androguard.core import apk  # Version >=4
    except ImportError:
        from androguard.core.bytecodes import apk  # Version < 4

    try:
        instance = apk.APK(path)
        instance.parse_v2_signing_block()
    except Exception:
        return ""

    def format_key(x):
        return "\n".join(textwrap.wrap(binascii.hexlify(x).decode("utf-8")))

    output = []
    for k, v in sorted(instance._v2_blocks.items()):
        output.append("Key {}:\n{}\n".format(hex(k), format_key(v)))

    return "\n".join(output)


def filter_apk_metadata(filepath, archive_name):
    new_filename = os.path.join(os.path.dirname(filepath), "APK metadata")

    logger.debug("Moving APK metadata from %s to %s", filepath, new_filename)

    # Strip the filename that was passed to apktool as its embedded in the
    # output. (It is unclear why this is conditional - see comments on
    # reproducible-builds/diffoscope#255)
    re_filename = re.compile(r"^apkFileName: %s" % re.escape(archive_name))

    with open(filepath) as in_, open(new_filename, "w") as out:
        out.writelines(x for x in in_ if not re_filename.match(x))

    os.remove(filepath)

    return new_filename



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import logging
import subprocess

from diffoscope.tools import tool_required


from .utils.file import File
from .utils.archive import Archive

logger = logging.getLogger(__name__)


class GzipContainer(Archive):
    def open_archive(self):
        return self

    def close_archive(self):
        pass

    def get_member_names(self):
        return [self.get_compressed_content_name(".gz")]

    @tool_required("gzip")
    def extract(self, member_name, dest_dir):
        dest_path = self.get_path_name(dest_dir)
        logger.debug("gzip extracting to %s", dest_path)
        with open(dest_path, "wb") as fp:
            # Use the short options (-d and -c) instead of the long options
            # (--decompress and --stdout) to support the gzip implementation
            # present in busybox, which doesn't support the long options.
            returncode = subprocess.call(
                ["gzip", "-d", "-c", self.source.path],
                stdout=fp,
                stderr=None,
            )
        # gzip exits with 2 (a warning) when decompression succeeds but there
        # is trailing garbage after the stream, eg. metadata appended to
        # OpenWrt firmware images. The payload is still fully decompressed, so
        # only treat a fatal error (exit code 1) as a failure.
        if returncode not in (0, 2):
            raise subprocess.CalledProcessError(
                returncode, ["gzip", "-d", "-c", self.source.path]
            )
        return dest_path


class GzipFile(File):
    DESCRIPTION = "Gzipped files"
    CONTAINER_CLASSES = [GzipContainer]
    FILE_TYPE_RE = re.compile(r"^gzip compressed data\b")

    # Work around file(1) Debian bug #876316
    FALLBACK_FILE_EXTENSION_SUFFIX = {".gz"}
    FALLBACK_FILE_TYPE_HEADER_PREFIX = b"\x1f\x8b"



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2021, 2025 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os
import re
import stat
import subprocess

from diffoscope.exc import RequiredToolNotFound
from diffoscope.tools import python_module_missing, tool_required
from diffoscope.config import Config
from diffoscope.difference import Difference

from .binary import FilesystemFile
from .missing_file import AbstractMissingType
from .utils.command import Command, our_check_output
from .utils.container import Container

logger = logging.getLogger(__name__)


if os.uname()[0] == "FreeBSD":

    class Stat(Command):
        @tool_required("stat")
        def cmdline(self):
            return [
                "stat",
                "-t",
                "%Y-%m-%d %H:%M:%S",
                "-f",
                "%Sp %l %Su %Sg %z %Sm %k %b %#Xf",
                self.path,
            ]

else:

    class Stat(Command):
        @tool_required("stat")
        def cmdline(self):
            return ["stat", self.path]

        FILE_RE = re.compile(r"^\s*File:.*$")
        DEVICE_RE = re.compile(r"Device: [0-9a-f]+h/[0-9]+d\s+")
        INODE_RE = re.compile(r"Inode: [0-9]+\s+")
        ACCESS_TIME_RE = re.compile(r"^Access: [0-9]{4}-[0-9]{2}-[0-9]{2}.*$")
        CHANGE_TIME_RE = re.compile(r"^Change: [0-9]{4}-[0-9]{2}-[0-9]{2}.*$")
        BIRTH_TIME_RE = re.compile(r"^\s*Birth:.*$")

        def filter(self, line):
            line = line.decode("utf-8")
            line = Stat.FILE_RE.sub("", line)
            line = Stat.DEVICE_RE.sub("", line)
            line = Stat.INODE_RE.sub("", line)
            line = Stat.ACCESS_TIME_RE.sub("", line)
            line = Stat.CHANGE_TIME_RE.sub("", line)
            line = Stat.BIRTH_TIME_RE.sub("", line)
            return line.encode("utf-8")


# compare only what matters
def stat_results_same(stat1, stat2):
    return stat_result_to_tuple(stat1) == stat_result_to_tuple(stat2)


def stat_result_to_tuple(s):
    return (
        s.st_mode,
        s.st_uid,
        s.st_gid,
        # Directory sizes are essentially arbitrary and filesystem dependent,
        # and may even differ, depending on the history of the directory.
        # (See MR !150)
        s.st_size if not stat.S_ISDIR(s.st_mode) else None,
        s.st_mtime,
    )


@tool_required("lsattr")
def lsattr(path):
    """
    NB. Difficult to replace with in-Python version. See
    <https://stackoverflow.com/questions/35501249/python-get-linux-file-immutable-attribute/38092961#38092961>
    """

    try:
        output = our_check_output(
            ["lsattr", "-d", path], stderr=subprocess.STDOUT
        ).decode("utf-8")
        return output.split()[0]
    except subprocess.CalledProcessError as e:
        if e.returncode == 1:
            # filesystem doesn't support xattrs
            return ""


class Getfacl(Command):
    @tool_required("getfacl")
    def cmdline(self):
        osname = os.uname()[0]
        if osname == "FreeBSD":
            return ["getfacl", "-q", "-h", self.path]
        return ["getfacl", "-p", "-c", self.path]


def xattr(path1, path2):
    try:
        import xattr as xattr_
    except ImportError:
        python_module_missing("xattr")
        return None

    # Support the case where the python3-xattr package is installed but
    # python3-pyxattr is not; python3-xattr has an xattr class that can be used
    # like a dict.
    try:
        get_all = xattr_.get_all
    except AttributeError:

        def get_all(x):
            return xattr_.xattr(x).items()

    def fn(x):
        return "\n".join(
            "{}: {}".format(
                k.decode("utf-8", "ignore"), v.decode("utf-8", "ignore")
            )
            for k, v in get_all(x)
        )

    return Difference.from_text(
        fn(path1), fn(path2), path1, path2, source="extended file attributes"
    )


def compare_meta(path1, path2):
    if Config().exclude_directory_metadata in ("yes", "recursive"):
        logger.debug(
            "Excluding directory metadata for paths (%s, %s)", path1, path2
        )
        return []

    logger.debug("compare_meta(%r, %r)", path1, path2)

    # Don't run any commands if any of the paths do not exist
    # or have other issues.
    try:
        stat1 = os.lstat(path1)
        stat2 = os.lstat(path2)
    except Exception:
        return []

    differences = []
    if stat_results_same(stat1, stat2):
        logger.debug("Stat structs are identical, moving on!")
    else:
        try:
            differences.append(
                Difference.from_operation(Stat, path1, path2, short=True)
            )
        except RequiredToolNotFound:
            logger.error("Unable to find 'stat'! Is PATH wrong?")

    if os.path.islink(path1) or os.path.islink(path2):
        return [d for d in differences if d is not None]

    if Config().extended_filesystem_attributes:
        try:
            differences.append(
                Difference.from_operation(Getfacl, path1, path2, short=True)
            )
        except RequiredToolNotFound:
            logger.info(
                "Unable to find 'getfacl', some directory metadata differences might not be noticed."
            )

        try:
            lsattr1 = lsattr(path1)
            lsattr2 = lsattr(path2)
            differences.append(
                Difference.from_text(
                    lsattr1, lsattr2, path1, path2, source="lsattr"
                )
            )
        except RequiredToolNotFound:
            logger.info(
                "Unable to find 'lsattr', some directory metadata differences might not be noticed."
            )
        differences.append(xattr(path1, path2))

    return [d for d in differences if d is not None]


def compare_directories(path1, path2, source=None):
    return FilesystemDirectory(path1).compare(FilesystemDirectory(path2))


class Directory:
    DESCRIPTION = "directories"

    @classmethod
    def recognizes(cls, file):
        return file.is_directory()

    @classmethod
    def fallback_recognizes(cls, file):
        return False


class FilesystemDirectory(Directory):
    def __init__(self, path):
        self._path = path

    @property
    def path(self):
        return self._path

    @property
    def name(self):
        return self._path

    @property
    def progress_name(self):
        x = self.name
        return x[1:] if x.startswith("./") else x

    @property
    def as_container(self):
        if not hasattr(self, "_as_container"):
            self._as_container = DirectoryContainer(self)
        return self._as_container

    def is_directory(self):
        return True

    def has_same_content_as(self, other):
        # no shortcut
        return False

    def compare(self, other, source=None):
        differences = []

        if not isinstance(other, FilesystemDirectory):
            return differences

        # We don't need to recurse into subdirectories; DirectoryContainer will
        # find them and do that for us.
        def list_files(path):
            return sorted(os.listdir(os.path.realpath(path)))

        listing_diff = Difference.from_text(
            "\n".join(list_files(self.path)),
            "\n".join(list_files(other.path)),
            self.path,
            other.path,
            source="file list",
        )
        if listing_diff:
            differences.append(listing_diff)

        if not isinstance(other, AbstractMissingType):
            differences.extend(compare_meta(self.name, other.name))

        my_container = DirectoryContainer(self)
        other_container = DirectoryContainer(other)
        differences.extend(my_container.compare(other_container))

        if not differences:
            return None

        difference = Difference(self.path, other.path, source)
        difference.add_details(differences)
        return difference


class DirectoryContainer(Container):
    def get_member_names(self):
        return sorted(os.listdir(self.source.path or "."))

    def get_member(self, member_name):
        member_path = os.path.join(self.source.path, member_name)

        if not os.path.islink(member_path) and os.path.isdir(member_path):
            return FilesystemDirectory(member_path)

        return FilesystemFile(
            os.path.join(self.source.path, member_name), container=self
        )



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2021-2022, 2024 Chris Lamb <lamby@debian.org>
# Copyright © 2021 Sergei Trofimovich
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import binascii
import io
import os
import re
import struct
import time

from diffoscope.difference import Difference

from .missing_file import MissingFile
from .utils.file import File

re_memory_address = re.compile(r" at 0x\w+(?=, )")


class PycFile(File):
    DESCRIPTION = "Python .pyc files"
    FILE_TYPE_RE = re.compile(
        r"(^python .*byte-compiled$|^Byte-compiled Python module for CPython .*)"
    )
    FALLBACK_FILE_EXTENSION_SUFFIX = {".pyc"}

    def compare_details(self, other, source=None):
        if isinstance(other, MissingFile):
            return []

        try:
            return [
                Difference.from_text(
                    describe_pyc(self.path),
                    describe_pyc(other.path),
                    self.path,
                    other.path,
                    source="Python bytecode",
                )
            ]
        except (ValueError, IndexError, struct.error) as exc:
            self.add_comment("Could not decompile bytecode: {}".format(exc))
            return []


def describe_pyc(filename):
    with open(filename, "rb") as f:
        return "\n".join(parse_pyc(f))


def parse_pyc(f):
    f.seek(0, io.SEEK_END)
    if f.tell() == 0:
        yield "type:     empty"
        return

    f.seek(0)
    magic = f.read(4)
    yield "magic:    {}".format(hexlify(magic))

    f.seek(4, 1)
    moddate = f.read(4)
    modtime = time.asctime(time.gmtime(struct.unpack("<L", moddate)[0]))
    yield "moddate:  {} ({} UTC)".format(hexlify(moddate), modtime)

    filesz = f.read(4)
    filesz = struct.unpack("<L", filesz)
    yield f"files sz: {filesz[0]}"

    start = f.tell()
    f.seek(0, os.SEEK_END)
    size = f.tell() - start
    yield f"code:     starts at offset {start} (size: {size} bytes)"


def hexlify(val):
    return "0x{}".format(binascii.hexlify(val).decode("utf-8"))



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016, 2018-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import os
import struct
import binascii

from diffoscope.difference import Difference

from .utils.file import File


class GitIndexFile(File):
    DESCRIPTION = "Git repositories"
    FILE_TYPE_RE = re.compile(r"^Git index")

    def compare_details(self, other, source=None):
        return [
            Difference.from_text(
                describe_index(self.path),
                describe_index(other.path),
                self.path,
                other.path,
            )
        ]


def parse_index(f):
    _, version = struct.unpack(">LL", f.read(4 * 2))

    return {"version": version, "entries": list(parse_entries(f))}


def parse_entries(f):
    num_entries = struct.unpack(">L", f.read(4))[0]

    for _ in range(num_entries):
        x = {}

        pos = f.tell()

        (
            x["ctime"],
            x["ctime_nano"],
            x["mtime"],
            x["mtime_nano"],
            x["dev"],
            x["inode"],
            x["mode"],
            x["uid"],
            x["gid"],
            x["size"],
            x["sha"],
            x["flags"],
        ) = struct.unpack(">LLLLLLLLLL20sH", f.read((4 * 10) + 20 + 2))

        x["path"] = f.read(x["flags"] & 0x0FFF)

        f.read((pos + ((f.tell() - pos + 8) & ~7)) - f.tell())

        yield x


def describe_index(filename):
    with open(filename, "rb") as f:
        index = parse_index(f)

    return """
Version: {version}

Entries:
{entries_fmt}
""".format(
        entries_fmt="".join(describe_entry(x) for x in index["entries"]),
        **index,
    )


def describe_entry(x):
    return """
Path:      {x[path]}
SHA:       {hexsha}
Size:      {x[size]}
Flags:     {x[flags]:#b}
User ID:   {x[uid]}
Group ID:  {x[gid]}
Created:   {x[ctime]}.{x[ctime_nano]}
Modified:  {x[mtime]}.{x[mtime_nano]}
Inode:     {x[inode]}
Device ID: ({major}, {minor})
""".format(
        x=x,
        major=os.major(x["dev"]),
        minor=os.minor(x["dev"]),
        hexsha=binascii.b2a_hex(x["sha"]).decode("utf-8"),
    )



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017-2021, 2024 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import re
import logging
import subprocess

from diffoscope.tools import tool_required, tool_check_installed
from diffoscope.tempfiles import get_temporary_directory
from diffoscope.difference import Difference

from .text import TextFile
from .utils.archive import Archive
from .utils.file import File
from .utils.command import Command, our_check_output

logger = logging.getLogger(__name__)

re_name = re.compile(r", created \d+, name=\"(?P<name>[^\"]+)\",")


class Pgpdump(Command):
    @tool_required("pgpdump")
    def cmdline(self):
        return (
            "pgpdump",
            "-i",  # Dump integer packets
            "-m",  # Dump marker packets
            "-p",  # Dump private packets
            "-u",  # Display UTC time
            self.path,
        )


class GpgListPackets(Command):
    @tool_required("gpg")
    def cmdline(self):
        return (
            "gpg",
            "--no-keyring",
            "--list-packets",
            self.path,
        )


class PGPContainer(Archive):
    @tool_required("gpg")
    def open_archive(self):
        # Extract to a fresh temporary directory.
        self._temp_dir = get_temporary_directory(suffix="pgp")

        try:
            our_check_output(
                (
                    "gpg",
                    "--decrypt",
                    "--no-keyring",
                    "--output",
                    os.path.join(self._temp_dir.name, "contents"),
                    os.path.abspath(self.source.path),
                ),
                cwd=self._temp_dir.name,
                stderr=subprocess.DEVNULL,
            )
        except subprocess.CalledProcessError:
            return False

        return self

    def close_archive(self):
        self._temp_dir.cleanup()

    def get_member_names(self):
        # Will only ever return one filename
        return os.listdir(self._temp_dir.name)

    def extract(self, member_name, dest_dir):
        return os.path.join(self._temp_dir.name, member_name)


class PgpFile(File):
    DESCRIPTION = "PGP signed/encrypted messages"
    FILE_TYPE_RE = re.compile(r"^(PGP message|PGP compressed data)\b")
    CONTAINER_CLASSES = [PGPContainer]
    FALLBACK_FILE_EXTENSION_SUFFIX = {".pgp", ".asc", ".pub", ".sec", ".gpg"}

    @classmethod
    def fallback_recognizes(cls, file):
        # Ensure we check FALLBACK_FILE_EXTENSION_SUFFIX, otherwise we run
        # pgpdump against all files that are recognised by file(1) as "data"
        if not super().fallback_recognizes(file):
            return False

        if file.magic_file_type == "data" and tool_check_installed("pgpdump"):
            try:
                output = our_check_output(
                    ("pgpdump", file.path), stderr=subprocess.DEVNULL
                )
            except subprocess.CalledProcessError:
                pass
            else:
                if b"New: unknown" not in output:
                    logger.debug("%s is a PGP file", file.path)
                    return True

            logger.debug("%s is not a PGP file", file.path)

        return False

    def compare_details(self, other, source=None):
        return [
            Difference.from_operation(
                Pgpdump, self.path, other.path, source="pgpdump"
            )
        ]


class PgpSignature(TextFile):
    DESCRIPTION = "PGP signatures"
    FILE_TYPE_RE = re.compile(r"^PGP signature\b")

    def compare(self, other, source=None):
        # Don't display signatures as hexdumps; use TextFile's comparisons...
        difference = super().compare(other, source)

        # ... but attach pgpdump of output
        difference.add_details(
            [
                Difference.from_operation(
                    Pgpdump, self.path, other.path, source="pgpdump"
                )
            ]
        )

        # ... as well as gpg --list-packets
        difference.add_details(
            [
                Difference.from_operation(
                    GpgListPackets,
                    self.path,
                    other.path,
                    source="gpg --list-packets",
                )
            ]
        )

        return difference
